@@ -51,7 +51,8 @@ RESOURCES = {
         'tenants': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-tenants.yml",
         'tokens': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-tokens.yml",
         'pgrest': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-pgrest.yml",
-        'jobs': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-jobs.yml"
+        'jobs': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-jobs.yml",
+        'apps': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-apps.yml"
     },
     'tapipy':{
         'actors': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-actors.yml',
@@ -121,8 +122,8 @@ def _get_specs(resources: Resources, spec_dir: str = None, download_latest_specs
     # Download and save specs if neccessary
     download_and_pickle_spec_dicts(resources, spec_dir=spec_dir, download_latest_specs=download_latest_specs)
     # Load, unpickle, and create specs
-    specs = unpickle_and_create_specs(resources, spec_dir=spec_dir)
-    return specs
+    specs, dicts = unpickle_and_create_specs(resources, spec_dir=spec_dir)
+    return specs, dicts
 
 
 def download_and_pickle_spec_dicts(resources: Resources, spec_dir: str, download_latest_specs: bool) -> None:
@@ -206,6 +207,7 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
     Can't be threaded, map doesn't allow spec object to be sent back.
     """
     specs = {}
+    dicts = {}
     # Get resource path to point the unpickling at.
     for resource_name, url in resources.items():
         if "local:" in url:
@@ -214,6 +216,7 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
                 with open(spec_path, 'rb') as spec_file:
                     spec_dict = yaml.load(spec_file, Loader=yaml.FullLoader)
                 specs.update({resource_name: create_spec(spec_dict)})
+                dicts.update({resource_name: spec_dict})
             except Exception as e:
                 print(f'Error reading local "{resource_name}" resource. '
                       f'Ensure path is absolute. e:{e}')
@@ -224,6 +227,7 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
             with open(spec_path, 'rb') as spec_file:
                 spec_dict = pickle.load(spec_file)
             specs.update({resource_name: create_spec(spec_dict)})
+            dicts.update({resource_name: spec_dict})
         except Exception as e:
             print(f'Got exception trying to unpickle and create spec for '
                   f'spec_path: "{resource_name}"; exception: {e}')
@@ -239,11 +243,12 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
                     with open(spec_path, 'rb') as spec_file:
                         spec_dict = pickle.load(spec_file)
                     specs.update({resource_name: create_spec(spec_dict)})
+                    dicts.update({resource_name: spec_dict})
                 except Exception as e:
                     print('Error opening tapipy prod spec. This is bad.')
             else:
                 print(f'No "{resource_name}" spec was found to fallback on')
-    return specs
+    return specs, dicts
 
 
 def update_spec_cache(resources: Resources = None, spec_dir: str = None) -> None:
@@ -254,16 +259,11 @@ def update_spec_cache(resources: Resources = None, spec_dir: str = None) -> None
     If a folder is specified, all urls specified are updated there.
     """
     if not resources:
-        # Get base resources from RESOURCES if resoruces not inputted
-        url_list = []
-        for resource_set in RESOURCES:
-            url_list.extend(list(RESOURCES[resource_set].values()))
-    else:
-        # Get just the URL's from the resources given
-        url_list = resources.values()
+        # Get base resources from RESOURCES if resources not inputted
+        resources = RESOURCES['tapipy']
 
     spec_dir = get_spec_dir(spec_dir)
-    download_and_pickle_spec_dicts(url_list, spec_dir=spec_dir, download_latest_specs=True)
+    download_and_pickle_spec_dicts(resources, spec_dir=spec_dir, download_latest_specs=True)
 
 
 def get_file_info_from_url(url: str, spec_dir: str):
@@ -302,7 +302,7 @@ def get_spec_dir(spec_dir: str):
     return spec_dir
     
 
-RESOURCE_SPECS = _get_specs(RESOURCES['tapipy'])
+RESOURCE_SPECS, RESOURCE_DICTS = _get_specs(RESOURCES['tapipy'])
 
 
 def get_basic_auth_header(username: str, password: str) -> str:
@@ -373,7 +373,8 @@ class Tapis(object):
                  download_latest_specs: bool = False,
                  spec_dir: str = None,
                  tenants: Tenants = None,
-                 is_tapis_service: bool = False
+                 is_tapis_service: bool = False,
+                 resource_dicts: dict = {}
                  ):
         # the base_url for the server this Tapis client should interact with
         self.base_url = base_url
@@ -393,9 +394,17 @@ class Tapis(object):
             self.account_type = 'user'
 
         # the access token to use -- should be an honest TapisResult access token.
+        if access_token and type(access_token) == str:
+            # if the access_token passed is a string, convert it to an honest TapisResult token:
+            tok = TapisResult(**{'access_token': access_token})
+            access_token = self.add_claims_to_token(tok)
         self.access_token = access_token
 
         # the refresh token to use -- should be an honest TapisResult refresh token.
+        if refresh_token and type(refresh_token) == str:
+            # if the access_token passed is a string, convert it to an honest TapisResult token:
+            tok = TapisResult(**{'refresh_token': refresh_token})
+            refresh_token = self.add_claims_to_token(tok)
         self.refresh_token = refresh_token
 
         # pass in a "raw" JWT directly. This is only used if the access_token is not set.
@@ -456,11 +465,11 @@ class Tapis(object):
         # Valuable so users don't overwrite their base specs.
         self.spec_dir = spec_dir
 
-        # Uses module instantiated RESOURCE_SPECS if there are no changes to the specs. 
+        # Uses module instantiated RESOURCE_SPECS if there are no changes to the specs.
         if self.custom_spec_dict or self.spec_dir or self.download_latest_specs or not self.resource_set == 'tapipy':
-            resource_specs = _get_specs(RESOURCES[resource_set], spec_dir=self.spec_dir, download_latest_specs=self.download_latest_specs)
+            resource_specs, self.resource_dicts = _get_specs(RESOURCES[resource_set], spec_dir=self.spec_dir, download_latest_specs=self.download_latest_specs)
         else:
-            resource_specs = RESOURCE_SPECS
+            resource_specs, self.resource_dicts = RESOURCE_SPECS, RESOURCE_DICTS
 
         # create resources for each API defined above. In the future we could make this more dynamic in multiple ways.
         for resource_name, spec in resource_specs.items():
@@ -651,20 +660,33 @@ class Tapis(object):
     def add_claims_to_token(self, token):
         """
         Adds claims and a callable expires_in() function to a Tapis token object.
-        :param tokne: (TapisResult) A TapisResult object returned using the t.tokens.create_token() method.
+        :param token: (TapisResult) A TapisResult object returned using the t.tokens.create_token() method.
         """
         def _expires_in():
-            return access_token.expires_at - datetime.datetime.now(datetime.timezone.utc)
+            return token_with_claims.expires_at - datetime.datetime.now(datetime.timezone.utc)
 
-        access_token = token
+        token_with_claims = token
+        # we could have been passed an access or a refresh token, so that will dictate which attribute is
+        # available:
+        try:
+            token_str = token_with_claims.access_token
+        except AttributeError:
+            token_str = token_with_claims.refresh_token
         # we do not validate the token because we may not have the cache of tenant objects (with corresponding
         # public keys) yet, in which case validation will fail
-        access_token.claims = jwt.decode(access_token.access_token, verify=False)
-        access_token.original_ttl = access_token.expires_in
-        access_token.expires_at = datetime.datetime.fromtimestamp(access_token.claims['exp'],
+        token_with_claims.claims = jwt.decode(token_str, verify=False)
+        # access_token.expires_in will not exist if the token is a string type... need to compute it from the
+        # claims...
+        if hasattr(token_with_claims, 'expires_in'):
+            token_with_claims.original_ttl = token_with_claims.expires_in
+        else:
+            # otherwise, we have no way of computing the original_ttl because all we have is the exp claim (an int)
+            # and the ttl would require also knowing when the token was created.
+            token_with_claims.original_ttl = -1
+        token_with_claims.expires_at = datetime.datetime.fromtimestamp(token_with_claims.claims['exp'],
                                                                   datetime.timezone.utc)
-        access_token.expires_in = _expires_in
-        return access_token
+        token_with_claims.expires_in = _expires_in
+        return token_with_claims
 
     def set_access_token(self, token):
         """
@@ -1179,6 +1201,8 @@ class Operation(object):
                                                    f"tenant {request_site_admin_tenant_id};")
             else:
                 pass
+        elif self.tapis_client.access_token:
+            access_token = self.tapis_client.access_token
         elif self.tapis_client.get_access_jwt():
             access_token = self.tapis_client.get_access_jwt()
         # if we got an access token, check if we need to refresh it
@@ -1195,7 +1219,9 @@ class Operation(object):
             if datetime.timedelta(seconds=5) > time_remaining:
                 # if the access token is about to expire, try to use refresh, unless this is a call to
                 # refresh (otherwise this would never terminate!)
-                if self.resource_name == 'tokens' and self.operation_id == 'refresh_token':
+                if (self.resource_name == 'tokens' and self.operation_id == 'refresh_token')\
+                        or (self.resource_name == 'authenticator' and self.operation_id == 'create_token'):
+
                     pass
                 else:
                     try:
